import java.io.*;
import java.net.*;

/**
 * This class creates a thread that listens to the output from the real robot.
 * It receives the joint coordinates and passes them on to the Cyclop-class.
 * @author Dieter Van Uytvanck
 * @version 0.1
 */
public class CyclopCommunicationReceiver extends Thread {
    
    private Cyclop refCyclop;
    private String address = "kweekvijver.elis.rug.ac.be";
    private final static int port = 1667;
    private Socket clientSocket;
    private DataInputStream isClient;
    private float baseCache, shoulderCache, elbowCache, wristTiltCache,
	wristTwistCache = 0;
    
    /** Creates a new instance of CyclopCommunicationReceiver and starts the 
     * listening thread.
     * @param c the Cyclop object to pass the received data to
     */
    public CyclopCommunicationReceiver(Cyclop c) {
        this.refCyclop = c;
        openSocket();
        // fire up the Thread's run() method
        this.start();
    }
    
    /** Opens the socket for listening. */
    private void openSocket() {
        try {
            clientSocket = new Socket(address, port);
    
	} catch (IOException e) {
            // Any IOException in big loop
            System.err.println("Exception while connecting"
            + " for connections: " + e);
        }
    }
    
    /** Closes the listening socket. */
    private void closeSocket() {
        try {
            clientSocket.close();
        }
        catch (UnknownHostException e) {
            System.err.println("Trying to connect to unknown host: " + e);
        }
        catch (IOException e) {
            System.err.println("IOException:  " + e);
        }
    }
    
    /** Starts the listening thread */
    public void run() {
        try {
	    isClient = new DataInputStream(clientSocket.getInputStream());
            
            float [] input = new float [6];
            int i = 0;
            
            while (true) {
                for (i=0; i<6; i++) {
                    // fill input buffer with all angles
                    input[i] = isClient.readFloat();
                }
                
		// process base angle
                if (input[0] != baseCache) {
                    refCyclop.setBaseAngle((double) input[0]);
                    baseCache = input[0];
                }
                
                // process shoulder angle
                if (input[1] != shoulderCache) {
                    refCyclop.setShoulderAngle((double) input[1]);
                    shoulderCache = input[1];
                }
                
                // process elbow angle
                if (input[2] != elbowCache) {
                    refCyclop.setElbowAngle((double) input[2]);
                    elbowCache = input[2];
                }
                
		// process WristTilt angle
                if (input[3] != wristTiltCache) {
                    refCyclop.setWristTiltAngle((double) input[3]);
                    wristTiltCache = input[3];
                }
                
		// process WristTwist angle
                if (input[4] != wristTwistCache) {
                    refCyclop.setWristTwistAngle((double) input[4]);
                    wristTwistCache = input[4];
                }
                
            }
        }
        
	// no input data available anymore
        catch (EOFException e) {
            // keep listening to socket input
            run();
        }
        // catch eventual errors
        catch (IOException e) {
            System.err.println("Couldn't get I/O for the connection to: hostname" + e);
        }
        // afterwards, always close the open sockets        
        finally {
            try {clientSocket.close(); closeSocket();}
            catch (Exception e) {System.err.println("Error while closing socket");}
        }
        
    }
    
}
