/**
 * JoystickInputDevice - This class serves as an InputDevice
 * in Java 3D which uses the Linux Joystick driver.
 */

package net.dixonpeugh.j3d.linux;

import javax.media.j3d.*;
import javax.vecmath.*;

import java.io.IOException;

public class JoystickInputDevice implements InputDevice {
    private Sensor joystickSensor;
    private SensorRead sensorRead = new SensorRead(4);
    private Transform3D newTransform = new Transform3D();

    private Transform3D rotTransX = new Transform3D();
    private Transform3D rotTransY = new Transform3D();
    private Transform3D rotTransZ = new Transform3D();

    private Vector3d initPos = new Vector3d();

    private String device;
    
    private int handle;
    private int currentMode = InputDevice.DEMAND_DRIVEN;
    
    private String joystickName;
    private int numAxis;
    private int numButtons;
    private int version;

    // Rotation Values
    /* .AxisRot - Index of the Axis value for Rotation. */
    private int xAxisRot = 0;
    private int yAxisRot = 1;
    private int zAxisRot = 2;

    /* .AxisFreedom - Radians of full rotation in one direction. */
    private double xAxisFreedom = Math.PI;
    private double yAxisFreedom = Math.PI;
    private double zAxisFreedom = Math.PI;

    // Translation Values
    /* .AxisTrans - Index of the Axis value for Translation */
    private int xAxisTrans = 0;
    private int yAxisTrans = 1;
    private int zAxisTrans = 2;

    /* .AxisDistance - Distance of the furthest point on the Axis. */
    private double xAxisDistance = 1.0;
    private double yAxisDistance = 1.0;
    private double zAxisDistance = 1.0;

    /* Other Constants */
    private double rotIncr = xAxisFreedom / 32767;
    private double transIncr = xAxisDistance / 32767;


    private int []axisData;
    private int []buttonData;
    
    static {
	System.loadLibrary("j3djoystick");
    }
    
    public JoystickInputDevice(String deviceName) {
	device = deviceName;
	initPos.set(0.0d, 0.0d, 0.0d);
    }

    public JoystickInputDevice() {
	device = "/dev/input/js0";
    }
    
    /* InputDevice Interface */
    public boolean initialize() {
	try {
	    handle = openDevice( device );
	    axisData = new int[numAxis];
	    buttonData = new int[numButtons];
	    
	    joystickSensor = new Sensor( this, 1, numButtons );
	    
	    return true;
	} catch (IOException e) {
	    e.printStackTrace();
	    return false;
	}
    }
    
    public void close() {
	closeDevice(handle);
    }
    
    public int getProcessingMode() {
	return currentMode;
    }
    
    public Sensor getSensor( int sensorIndex ) {
	return joystickSensor;
    }
    
    public int getSensorCount() {
	if (joystickSensor == null) return 0;
	return 1;
    }
    
    public void pollAndProcessInput() {
	try {
	    pollDevice(handle, axisData, buttonData);

	    sensorRead.setTime( System.currentTimeMillis() );

	    rotTransX.rotX( axisData[xAxisRot] * rotIncr );
	    rotTransY.rotY( axisData[yAxisRot] * rotIncr );
	    rotTransZ.rotZ( axisData[zAxisRot] * rotIncr );

	    Vector3d newPos = new Vector3d ( axisData[xAxisTrans] * transIncr,
					     axisData[yAxisTrans] * transIncr,
					     axisData[zAxisTrans] * transIncr);
	    newPos.add( initPos );
	    newTransform.set( newPos );

	    newTransform.mul( rotTransX );
	    newTransform.mul( rotTransY );
	    newTransform.mul( rotTransZ );

	    sensorRead.set( newTransform );
	    sensorRead.setButtons ( buttonData );
	    joystickSensor.setNextSensorRead( sensorRead );
						  
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
    
    public void processStreamInput() {
	// No Implementation required.
    }
    
    public void setNominalPositionAndOrientation() {
	
    }
    
    public void setProcessingMode( int mode ) {
	if (mode == currentMode) return;
	switch (mode) {
	case BLOCKING:
	    setDeviceMode(handle, 0);
	    currentMode = mode;
	    break;
	case NON_BLOCKING:
	case DEMAND_DRIVEN:
	    setDeviceMode(handle, 1);
	    currentMode = mode;
	    break;
	}  
    }
    
    
    /* Native Methods */
    private native int openDevice( String deviceName )
	throws IOException;
    private native void closeDevice(int handle);
    
    private native void pollDevice(int handle, 
				   int[] axisData, 
				   int[] buttonData)
	throws IOException;
    private native void setDeviceMode(int handle, int mode);
    
    
    /* Informative Methods */
    public String getName() {
	return joystickName;
    }
    
    public int getNumButtons() {
	return numButtons;
    }
    
    public int getNumAxis() {
	return numAxis;
    }
    
    public long getVersion() {
	return version;
    }
    
    public static void main(String args[]) {
	JoystickInputDevice jid = new JoystickInputDevice("/dev/input/js0");
	
	jid.initialize();
	
	System.out.println("Joystick Information:");
	System.out.println("\tName: " + jid.getName());
	System.out.print("\t# Buttons: ");
	System.out.println(jid.getNumButtons());
	
	System.out.print("\t# Axes: " );
	System.out.println(jid.getNumAxis());
	
	System.out.print("\tVersion: ");
	System.out.print(jid.getVersion() >> 16);
	
	System.out.print(".");
	System.out.print((jid.getVersion() >> 8) & 0xff);
	System.out.print(".");
	System.out.println(jid.getVersion() & 0xff);
	
	jid.setProcessingMode(NON_BLOCKING);
	try {
	    int i = 0;
	    while (true) {
		jid.pollAndProcessInput();
		System.out.print(i++);
		for (int j = 0; j < jid.axisData.length; j++) {
		    System.out.print("\tA" + Integer.toString(j) + ": " +
				     Integer.toString(jid.axisData[j]));
		    
		}
		for (int k = 0; k < jid.buttonData.length; k++) {
		    System.out.print("\tB" + Integer.toString(k) + ": " +
				     Integer.toString(jid.buttonData[k]));
		}
		System.out.print("\r");
		
		System.out.flush();
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
	
	jid.close();
    }
}
